﻿Option Explicit On
Option Strict On
#If VBC_VER >= 9 Then
Option Infer Off
#End If

Imports System
Imports System.Runtime.InteropServices
Imports System.ComponentModel
Imports System.Reflection

<Assembly: AssemblyTitle("VDotNetLib.NativeLibrary")> 
<Assembly: AssemblyDescription("This library can be used to load native libraries and get the address of functions inside.")> 
<Assembly: AssemblyProduct("VDotNetLib.NativeLibrary")> 
<Assembly: AssemblyCopyright("Copyright © Michaël De Vos 2010")> 
<Assembly: CLSCompliant(True)> 
<Assembly: ComVisible(False)> 
<Assembly: Guid("e9348253-f7bf-481d-8caf-5fd29362f194")> 
<Assembly: AssemblyVersion("1.0.0")> 

Namespace VDotNetLib

    ''' <summary>
    ''' Class that holds a native library. This class can be used to load native libraries and get the address of functions inside.
    ''' </summary>
    ''' <remarks>When the class is disposed or garbage collected and the library was loaded by LoadLibrary, FreeLibrary is called.</remarks>
    Public Class NativeLibrary
        Implements IDisposable

        Private _lib As IntPtr
        Private _mustFree As Boolean

        Private Sub New()
            Me._mustFree = False
            Me._lib = IntPtr.Zero
        End Sub

        Protected Overrides Sub Finalize()
            Me.Free(False)
            MyBase.Finalize()
        End Sub

        ''' <summary>
        ''' Tries to load the library first with GetModuleHandle. If this fails 
        ''' the methods tries again with LoadLibrary. After loading it is possible
        ''' to call GetProcAddress to get an address to a function in the loaded library.
        ''' </summary>
        ''' <param name="fileName">The name or location of the library. See library search order...</param>
        ''' <returns>A instance of NativeLibrary.</returns>
        ''' <remarks>Throws an exception on failure.</remarks>
        Public Overloads Shared Function Load(ByVal fileName As String) As NativeLibrary
            Return Load(fileName, True)
        End Function

        ''' <summary>
        ''' See <see cref="Load">Load</see>.
        ''' </summary>
        ''' <param name="fileName">See <see cref="Load">Load</see>.</param>
        ''' <returns>See <see cref="Load">Load</see>.</returns>
        ''' <remarks>Throws no exception on failure.</remarks>
        Public Overloads Shared Function TryLoad(ByVal fileName As String) As NativeLibrary
            Return Load(fileName, False)
        End Function

        ''' <summary>
        ''' Property that returns a value that indicates if the library must be freed. (If True the library was loaded by LoadLibrary and not GetModuleHandle.)
        ''' </summary>
        ''' <returns>True when the library must be freed. False if not.</returns>
        Public ReadOnly Property MustFree() As Boolean
            Get
                Return Me._mustFree
            End Get
        End Property

        ''' <summary>
        ''' Explicit frees the loaded library. Does not free the object.
        ''' </summary>
        ''' <remarks>
        ''' It is not neciserry to explicitly free the loaded library. This is done automatically when the object gets garbage collected. 
        ''' The dispose method does the same... 
        ''' This method throws an exception on failure.
        ''' </remarks>
        Public Sub Free()
            Me.Free(True)
        End Sub

        ''' <summary>
        ''' Explicit frees the loaded library. Does not free the object.
        ''' </summary>
        ''' <remarks>
        ''' See <see cref="Free">Free</see>. 
        ''' This method throws no exception on failure.
        ''' </remarks>
        Public Sub TryFree()
            Me.Free(False)
        End Sub

        ''' <summary>
        ''' This method gets the address for a method inside the loaded library.
        ''' </summary>
        ''' <param name="procName">The name of the method.</param>
        ''' <returns>The address of the method.</returns>
        ''' <remarks>Throws an exception on failure.</remarks>
        Public Function GetProcAddress(ByVal procName As String) As IntPtr
            Return GetProcAddress(procName, True)
        End Function

        ''' <summary>
        ''' See <see cref="GetProcAddress">GetProcAddress</see>
        ''' </summary>
        ''' <param name="procName">The name of the method.</param>
        ''' <returns>The address of the method.</returns>
        ''' <remarks>Throws no exception on failure.</remarks>
        Public Function TryGetProcAddress(ByVal procName As String) As IntPtr
            Return GetProcAddress(procName, False)
        End Function

#Region " Private Methods "

        Private Overloads Shared Function Load(ByVal fileName As String, ByVal throwOnError As Boolean) As NativeLibrary
            Dim nl As New NativeLibrary
            nl._lib = NativeMethods.GetModuleHandle(fileName)
#If NET10 Then
            If nl._lib.ToInt32() = 0 Then
#Else
            If nl._lib = IntPtr.Zero Then
#End If
                nl._lib = NativeMethods.LoadLibrary(fileName)
#If NET10 Then
                If nl._lib.ToInt32() = 0 Then
#Else
                If nl._lib = IntPtr.Zero Then
#End If
                    If throwOnError Then
                        Dim w32err As Integer = Marshal.GetLastWin32Error()
                        If w32err = 126 Then
                            Throw New ComponentModel.Win32Exception(w32err, "The specified module could not be found.")
                        Else
                            Throw New ComponentModel.Win32Exception(w32err)
                        End If
                    Else
                        Return Nothing
                    End If
                Else
                    nl._mustFree = True
                End If
            End If
            Return nl
        End Function

        Private Function GetProcAddress(ByVal procName As String, ByVal throwOnError As Boolean) As IntPtr
            Dim p As IntPtr = NativeMethods.GetProcAddress(Me._lib, procName)
#If NET10 Then
            If p.ToInt32() = 0 Then
#Else
            If p = IntPtr.Zero Then
#End If
                If throwOnError Then
                    Throw New ComponentModel.Win32Exception()
                Else
                    Return IntPtr.Zero
                End If
            Else
                Return p
            End If
        End Function

        Private Sub Free(ByVal throwOnError As Boolean)
#If NET10 Then
            If Me._mustFree AndAlso Me._lib.ToInt32() <> 0 Then
                If NativeMethods.FreeLibrary(Me._lib) = 0 AndAlso throwOnError Then
                    Throw New ComponentModel.Win32Exception()
                End If
                Me._mustFree = False
                Me._lib = IntPtr.Zero
            End If
#Else
            If Me._mustFree AndAlso Me._lib <> IntPtr.Zero Then
                If Not NativeMethods.FreeLibrary(Me._lib) AndAlso throwOnError Then
                    Throw New ComponentModel.Win32Exception()
                End If
                Me._mustFree = False
                Me._lib = IntPtr.Zero
            End If
#End If
        End Sub

        Private NotInheritable Class NativeMethods

            Private Const COREDLL_DLL As String = "Coredll.dll"

            <DllImport(COREDLL_DLL, SetLastError:=True)> _
            Friend Shared Function GetModuleHandle(ByVal fileName As String) As IntPtr
            End Function

            <DllImport(COREDLL_DLL, SetLastError:=True)> _
            Friend Shared Function LoadLibrary(ByVal fileName As String) As IntPtr
            End Function

#If NET10 Then
            <DllImport(COREDLL_DLL, SetLastError:=True)> _
            Friend Shared Function FreeLibrary(ByVal library As IntPtr) As Integer
            End Function
#Else
            <DllImport(COREDLL_DLL, SetLastError:=True)> _
            Friend Shared Function FreeLibrary(ByVal library As IntPtr) As <MarshalAs(UnmanagedType.Bool)> Boolean
            End Function
#End If
            <DllImport(COREDLL_DLL, SetLastError:=True)> _
            Friend Shared Function GetProcAddress(ByVal library As IntPtr, ByVal procName As String) As IntPtr
            End Function

            Private Sub New()
            End Sub

        End Class

#End Region

#Region " IDisposable Support "
        Private disposedValue As Boolean = False
        Protected Overridable Sub Dispose(ByVal disposing As Boolean)
            If Not Me.disposedValue Then
                Me.disposedValue = True
                Me.Free(False)
            End If
        End Sub
        Public Sub Dispose() Implements IDisposable.Dispose
            Dispose(True)
            GC.SuppressFinalize(Me)
        End Sub
#End Region

    End Class

End Namespace
